/**
seglocStatsIO.c : Functions to save (writeDataFile) and re-read
                  (readDataFile) all data to/from files that were used
		  to localized the robot : parameters, position of the robot,
                  number of acquired segments, number of segments of the
                  aspect map, list of the solutions with, for each one,
                  the list of the matching segments.
                  
                  Produces its results in file with '-segloc.nnn'
                  extension name (nnn is an incremented number).
                  See for example the file:  example-segloc.001

                  Very usefull for debug purpose and also to analysis the
                  quality of the localisation using filter-stats.c
                  functions (SEE KARIM FAKHFAKH WORK).

                   TO RECORD THOSE DATA, COMPILE THE MODULE WITH FLAG:
                   'CLASSIFICATION'
**/

#ifdef VXWORKS
#include "vxWorks.h"
#endif
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dbliste.h>
#include <lasdef.h>
#include <las_struct_p.h>
/* #include "las_prot.h" */
/* #include "image_mgr_p.h" */
#include <las_io_p.h>
#include <map_util.h>
#include <localize.h>
#include <print_results_p.h>

#include "pomStruct.h"
#include "sickStruct.h"
#include "seglocStruct.h"
#include "seglocStatsIO.h"
#include <h2mathLib.h>

static char *seglocClassName[] = SEGLOC_CLASS_NAME;

#define GAUSS_99 3.0 

/*----------------------------------------------------------------------
 *
 * writeDataFile  -  save segloc results in file
 *
 *
 *
 */
int writeDataFile(FILE *file, SEGLOC_SEG_PARAMS *segParams,
		  SEGLOC_LOC_MODE locMode,
		  SEGLOC_LOC_PARAMS *locParams,
		  SEGLOC_POS_VAR *initPos,
		  SICK_MEASURES_HEADER_STR *header,
		  SEGLOC_HYPS hypTbl[],
		  int nbSol, int best,
		  int nbAcqSegs, double acqLength, double acqAngleView,
		  int nbAspectSegs, double aspectLength, double aspectAngleView)
{
  char buf[64];
  int i;
  Hyp *hyp;

  /* Segmentation params */
  fprintf(file, "segs: thres %g min_pts %d max_dist %g\n",
	  segParams->leastSquareThres,
	  segParams->nbMinSegPts,
	  segParams->distMaxSegPts);
  
  /* Localisation params */
  switch (locMode) {
  case SEGLOC_LOC_LOCAL:
    sprintf(buf, "local");
    break;
  case SEGLOC_LOC_ONCE:
    sprintf(buf, "once");
    break;
  case SEGLOC_LOC_LOOP:
    sprintf(buf, "loop");
    break;
  case SEGLOC_FORCE_LOC_LOCAL:
    sprintf(buf, "force");
    break;
  }
  
  fprintf(file, 
	  "loc: %s  Fvar %c Fvisi %c posMaha %ld mapMaha %ld \n"
	  "     angleUnc %.3f factUnc %5.2f\n",
	  buf, 
	  locParams->variancesFilter ? 'T' : 'F',
	  locParams->visibilityFilter ? 'T' : 'F',
	  locParams->posDistMahaThres,
	  locParams->mapDistMahaThres,

	  locParams->angleUncertainty,
	  locParams->factUncertaintyTest);
  
  /* Current pos and co */
  fprintf(file,
  	  "robot     x %.3f y %.3f th %.3f  vx %.3f vy %.3f w %.3f \n"
	  "  var %.5f %.5f %.5f %.5f %.5f %.5f\n",
	  
 	  header->sickPomPos.mainToBase.euler.x,
	  header->sickPomPos.mainToBase.euler.y,
	  header->sickPomPos.mainToBase.euler.yaw,
	  
	  header->sickPomPos.VLocal.euler.x,
	  header->sickPomPos.VLocal.euler.y,
	  header->sickPomPos.VLocal.euler.yaw,

	  header->sickPomPos.mainToBase.var.vxx,
	  header->sickPomPos.mainToBase.var.vxy,
	  header->sickPomPos.mainToBase.var.vyy, 
	  header->sickPomPos.mainToBase.var.vox, 
	  header->sickPomPos.mainToBase.var.voy,
	  header->sickPomPos.mainToBase.var.voo);
  
  /* Init pos for loc */
  fprintf(file,
  	  "init pos  x %.3f y %.3f th %.3f\n"
	  "  var %.5f %.5f %.5f %.5f %.5f %.5f\n",
	  
	  initPos->pos.x, initPos->pos.y, initPos->pos.theta,

	  initPos->var.vxx, initPos->var.vxy, initPos->var.vyy, 
	  initPos->var.vxt, initPos->var.vyt, initPos->var.vtt);

  /* Segmentation results */
  fprintf(file, "acq    segs: %d length %5.2f angle %5.2f\n",
	  nbAcqSegs, acqLength, acqAngleView);
  fprintf(file, "aspect segs: %d length %5.2f angle %5.2f\n",
	  nbAspectSegs, aspectLength, aspectAngleView);
   
  /* Nb solutions */
  fprintf(file,
	  "nb sol %d  best %d\n", nbSol, best);
  

  /** 
   ** Record each hyp in file
   **/
  for (i=0; i<nbSol; i++) {
    hyp = hypTbl[i].hyp;
    fprintf (file, "%-2d pos  (%-5.2f %-5.2f %-5.2f) nm %-2ld lm %-.2g %s %c\n",
	     i,
	     hyp->Tr[0], hyp->Tr[1], hyp->Tr[2], 
	     hyp->nmatched, hyp->length_matched, 
	     seglocClassName[hypTbl[i].class], i==best?'*':'.');
    
    fprintf (file, "   vpos (%.5f %.5f %.5f %.5f %.5f %.5f)\n",
	     hyp->Vtr[0], hyp->Vtr[1], hyp->Vtr[2], 
	     hyp->Vtr[3], hyp->Vtr[4], hyp->Vtr[5]);
    
    print_hyp_segs(file, hyp->lsm, hyp->Tr, hyp->Vtr); 
  }
  
  return TRUE;
}


/*----------------------------------------------------------------------
 *
 * readDataFile  -  read segloc results from file
 *
 *
 *
 */
int readDataFile(FILE *pfilein, SEGLOC_SEG_PARAMS *segParams,
		 SEGLOC_LOC_MODE *locMode,
		 SEGLOC_LOC_PARAMS *locParams,
		 SEGLOC_POS_VAR *initPos,
		 SICK_MEASURES_HEADER_STR *header,
		 SEGLOC_HYPS_STR *hypTbl[],
		 int *nbAcqSegs, double *acqLength, double *acqAngleView,
		 int *nbAspectSegs, double *aspectLength, double *aspectAngleView)
{			
  int nbscan;
  int numhyp, nummatch, num;
  Hyp *hyp;

  Sgl *sp;  /* Le segment "percu" du matching */
  Sgl *sm;  /* Le segment "model" du matching */

  int nbSol;
  int best;
  char locmode[16];
  char variancesFilterChar;
  char visibilityFilterChar;
  char classname[16];
  char bestchar;
		 
  /* Segmentation params */
  if (fscanf(pfilein, 
	     "segs: thres %lf min_pts %d max_dist %lf\n",
	     &segParams->leastSquareThres,
	     &segParams->nbMinSegPts,
	     &segParams->distMaxSegPts) != 3) {
    printf ("Cannot read segs params...\n");
    return 0;
  }

  /* Lecture en-tete 1 (parametres loc) */
  if (fscanf(pfilein, 
	     "loc: %s  Fvar %c Fvisi %c posMaha %ld mapMaha %ld \n"
	     "     angleUnc %lf factUnc %lf\n",
	     locmode, 
	     &variancesFilterChar,
	     &visibilityFilterChar,
	     &locParams->posDistMahaThres,
	     &locParams->mapDistMahaThres,

	     &locParams->angleUncertainty,
	     &locParams->factUncertaintyTest) != 7) {
    printf ("Cannot read loc ...\n");
    return 0;
  }
  
  if (strcmp(locmode, "local")==0)
    *locMode = SEGLOC_LOC_LOCAL;
  else if(strcmp(locmode, "once")==0)
    *locMode = SEGLOC_LOC_ONCE;
  else if(strcmp(locmode, "loop")==0)
    *locMode = SEGLOC_LOC_LOOP;

  if (variancesFilterChar == 'T')
    locParams->variancesFilter  = 1;
  else
    locParams->variancesFilter  = 1;
    
  if (visibilityFilterChar == 'T')
    locParams->visibilityFilter  = 1;
  else
    locParams->visibilityFilter  = 1;
  
    
  /* Lecture en-tte 2 (Current pos and co) */
  if (fscanf(pfilein,
	     "robot x %lf y %lf th %lf  vx %lf vy %lf w %lf \n"
	     "  var %lf %lf %lf %lf %lf %lf\n",

	     &header->sickPomPos.mainToBase.euler.x,
	     &header->sickPomPos.mainToBase.euler.y,
	     &header->sickPomPos.mainToBase.euler.yaw,
	     
	     &header->sickPomPos.VLocal.euler.x,
	     &header->sickPomPos.VLocal.euler.y,
	     &header->sickPomPos.VLocal.euler.yaw,
	     
	     &header->sickPomPos.mainToBase.var.vxx,
	     &header->sickPomPos.mainToBase.var.vxy, 
	     &header->sickPomPos.mainToBase.var.vyy, 
	     &header->sickPomPos.mainToBase.var.vox, 
	     &header->sickPomPos.mainToBase.var.voy,
	     &header->sickPomPos.mainToBase.var.voo)!= 11) {
    printf ("Cannot read robot ...\n");
    return 0;
  }
  
  /* Init pos for loc */
  if (fscanf(pfilein,
	     "init pos  x %lf y %lf th %lf\n"
	     "  var %lf %lf %lf %lf %lf %lf\n",
	  
	     &initPos->pos.x, &initPos->pos.y, &initPos->pos.theta,
	     
	     &initPos->var.vxx, &initPos->var.vxy, &initPos->var.vyy, 
	     &initPos->var.vxt, &initPos->var.vyt, &initPos->var.vtt) != 9) {
    printf ("Cannot read init pos ...\n");
    return 0;
  }
  
      
  /* Segmentation results */
  if (fscanf(pfilein,
	     "acq    segs: %d length %lf angle %lf\n",
	     nbAcqSegs, acqLength, acqAngleView) != 3) {
    printf ("Cannot read acq segs ...\n");
    return 0;
  }
  
  if (fscanf(pfilein,
	     "aspect segs: %d length %lf angle %lf\n",
	     nbAspectSegs, aspectLength, aspectAngleView) != 3) {
    printf ("Cannot read aspect segs ...\n");
    return 0;
  }

  /* Lecture en-tte 3 (les solutions) */
  if (fscanf(pfilein,
	     "nb sol %d  best %d\n", &nbSol, &best) != 2) {
    printf ("Cannot read nbsol ...\n");
    return 0;
  }
  
  /* Lecture solutions */
  if ((*hypTbl = malloc(sizeof(SEGLOC_HYPS_STR) * nbSol)) == NULL) {
    printf ("Cannot alloc hypTbl ...\n");
    return 0;
  }
  
  for(numhyp=0; numhyp<nbSol; numhyp++) {
    
    hyp = &((*hypTbl)[numhyp].hyp);
    
    nbscan = fscanf (pfilein, 
		     "%d pos  (%lf %lf %lf ) nm %ld lm %lf %s %c",
		     &num,
		     &hyp->Tr[0], &hyp->Tr[1], &hyp->Tr[2], 
		     &hyp->nmatched, &hyp->length_matched, 
		     classname, &bestchar);
    if (nbscan != 8) {
      printf ("Cannot read hyp %d ...\n", numhyp);
      return 0;
    }
    if (bestchar == '*') {
      (*hypTbl)[numhyp].bestsegloc = 1;
    }
    else
      (*hypTbl)[numhyp].bestsegloc = 0;
    
    (*hypTbl)[numhyp].class = -1;
    for (num=0; num<4; num++) {
      if (strcmp(classname, seglocClassName[num]) == 0) {
	(*hypTbl)[numhyp].class = num;
	break;
      }
    }
    if ((*hypTbl)[numhyp].class == -1) {
      printf ("Unknow class hyp %d\n", numhyp);
      return 0;
    }
    
    nbscan = fscanf (pfilein, " vpos (%lf %lf %lf %lf %lf %lf) \n",
		     &hyp->Vtr[0], &hyp->Vtr[1], &hyp->Vtr[2], 
		     &hyp->Vtr[3], &hyp->Vtr[4], &hyp->Vtr[5]);
    if (nbscan != 6) {
      printf ("Cannot read hyp %d ux ... (nbscan %d)\n", numhyp, nbscan);
      return 0;
    }
        
    /* Lecture des segments */
    if (( (*hypTbl)[numhyp].segmod = 
	  malloc(sizeof(Sgl) * hyp->nmatched)) == NULL) {
      printf ("Cannot alloc  segmod\n");
      return 0;
    }
    if (( (*hypTbl)[numhyp].segacq = 
	malloc(sizeof(Sgl) * hyp->nmatched)) == NULL) {
      printf ("Cannot alloc  segmod\n");
      return 0;
    }
    
    if (( (*hypTbl)[numhyp].robotAngle = 
	malloc(sizeof(double) * hyp->nmatched)) == NULL) {
      printf ("Cannot alloc  robotAngle\n");
      return 0;
    }
    
    if (( (*hypTbl)[numhyp].numgp = 
	  malloc(sizeof(int) * hyp->nmatched)) == NULL) {
      printf ("Cannot alloc  numgp\n");
      return 0;
    }

    for(nummatch=0; nummatch<hyp->nmatched; nummatch++) {
      
      if (fscanf (pfilein, "  %d ", &num) != 1) {
	printf ("Cannot read hyp %d segs %d ...\n", numhyp, nummatch);
	return 0;
      }
      
      sm = &((*hypTbl)[numhyp].segmod)[nummatch];
      if (fscanf (pfilein, "    m %ld : (%lf %lf) (%lf %lf)   vo %lf",
		  &sm->ifirst, &sm->first.x, &sm->first.y, &sm->last.x,
		  &sm->last.y, &sm->vo) != 6) {
	printf ("Cannot read sm hyp %d seg %d ...\n", numhyp, nummatch);
	return 0;
      }
      
      if (fscanf (pfilein, " ang %lf  gpe %d \n", 
		  &((*hypTbl)[numhyp].robotAngle)[nummatch],
		  &((*hypTbl)[numhyp].numgp)[nummatch]) != 2) {
	printf ("Cannot read ang hyp %d segs %d ...\n", numhyp, nummatch);
	return 0;
      }
      
      sp = &((*hypTbl)[numhyp].segacq)[nummatch];
      if (fscanf (pfilein, "         p %ld : (%lf %lf) (%lf %lf)   vo %lf\n",
		  &sp->ifirst, &sp->first.x, &sp->first.y, &sp->last.x, 
		  &sp->last.y, &sp->vo) != 6) {
	printf ("Cannot read sp hyp %d seg %d ...\n", numhyp, nummatch);
	return 0;
      }
      
    } /* FOR Segments */
  } /* FOR Sols */
  return nbSol;
}

