/**
 ** Copyright (c) 2000 LAAS/CNRS 
 **
 ** Sara Fleury - 
 **
 ** Useful functions to compute the intersection of polygons and segments.
 ** 
 ** 
 **/
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <portLib.h>
#include <h2mathLib.h>

#include "polygons.h"

/**
 ** A point is described as an array of 2 double.
 **
 ** A polygon is described by :
 **
 **     - an int 'nbPts' that indicates the number of points.
 **
 **     - an array 'poly' of '2*nbPts' double with :
 **
 **                poly[0]    poly[1]     : 1-st point coord
 **                poly[2]    poly[3]     : 2-nd point coord
 **                ...
 **                poly[2*i]  poly[2*i+1] : i-th point coord
 **
 **   Polygons can be closed or open (ie, 1st and last points can 
 **   be identical).
 **
 **   For most functions, like intersection or belonging, their points 
 **   can be given in trigo or reverse order (does not depend on matter 
 **   side) BUT for the polygon growing :
 **   the growing is done on the left side if "walking" from point 1 
 **   to 2, in other words, the polygon is grown if its points are 
 **   given in clock
 **   order or reduced if given in trigo order. However, as you never know
 **   how your polygons will be used in futur we recommend you to allways
 **   defined them in same order. 
 **   You have to think that the MATTER SIDE, if any, is on the RIGHT
 **   when walking from point 1 to 2.
 **   THUS : 
 **   - polygon that represent OBSTACLES should be given CLOCK wise
 **   - polygon that represent FREE SPACE should be given TRIGO wise
 **/

/*----------------------------------------------------------------------
 *  PLpolygonInterPolygon   - 
 *
 * Description : Test polygon/polygon surfaces intersection
 * Input       : 2 polynoms
 * Returns     : TRUE if intersection (including the border), FALSE otherwise
 */
int PLpolygonInterPolygon (int nbPts1, double *poly1, 
			   int nbPts2, double *poly2)
{
  int i, j;
  double xMinR, yMinR, xMaxR, yMaxR;
  double xMinP, yMinP, xMaxP, yMaxP;
  
  /* Polygones ferms ? */
  if (PLtestClosedPolygon(nbPts1, poly1))
    nbPts1--;
  if (PLtestClosedPolygon(nbPts2, poly2))
    nbPts2--;

  /* Test validite */
  if (nbPts2 < 1 || nbPts1 < 1 || (nbPts1==1 && nbPts2==1)) {
    printf ("PLpolygonInterPolygon: unvalid parameters\n");
    return FALSE;
  }

  /* Polygone a un point !? */
  if (nbPts2 == 1)
    return (PLpointInPolygon(poly2, nbPts1, poly1));
  if (nbPts1 == 1)
    return (PLpointInPolygon(poly1, nbPts2, poly2));

  /* Rectangles ex-inscrits aux polygones */
  PLpolygonMinMax(nbPts1, poly1, &xMinR, &yMinR, &xMaxR, &yMaxR);
  PLpolygonMinMax(nbPts2, poly2, &xMinP, &yMinP, &xMaxP, &yMaxP);

  /* Tester si les rectangles ex-inscrits s'intersectent */
  if ((xMinR > xMaxP) || (yMinR > yMaxP)
      || (xMaxR < xMinP) || (yMaxR < yMinP))
    return(FALSE);
  
  /* Regarder si un sommet du polygone1 est dans le polygone2 */
  if (nbPts2 > 2)
    for (i=0; i<nbPts1; i++) 
      if (PLpointInPolygon(&poly1[2*i], nbPts2, poly2))
	return(TRUE);
  /* Tester l'inverse */
  if (nbPts1 > 2)
    for (i=0; i<nbPts2; i++) 
      if (PLpointInPolygon(&poly2[2*i], nbPts1, poly1))
	return(TRUE);
  
  /* Regarder l'intersection des segments */
  for (i=0; i<nbPts1-1; i++) {
    for (j=0; j<nbPts2-1; j++) 
      if (PLtestIntersectSegments(&poly1[2*i], &poly1[2*(i+1)], &poly2[2*j], &poly2[2*(j+1)]))
	return(TRUE);
    if (PLtestIntersectSegments(&poly1[2*i], &poly1[2*(i+1)], &poly2[2*j], &poly2[0]))
	return(TRUE);
  }
  for (j=0; j<nbPts2-1; j++) 
    if (PLtestIntersectSegments(&poly1[2*i], &poly1[0], &poly2[2*j], &poly2[2*(j+1)]))
      return(TRUE);
  if (PLtestIntersectSegments(&poly1[2*i], &poly1[0], &poly2[2*j], &poly2[0]))
    return(TRUE);
  
  return(FALSE);
	  
}


/*----------------------------------------------------------------------
 * PLrectangleInterPolygon   - 
 *
 * Description : Test if a rectangle intersect a polygon
 * Input       : 1 polynom with 4 vertices and 1 general polygon
 * Returns     : TRUE if intersection (including the border), FALSE otherwise
 *               WARNING: FALSE if the polygon is entirely inside the rectangle
 */
int PLrectangleInterPolygon (double rectangle[2*4], int nbPts, double *poly)
{
  int i, j;
  double xMinR, yMinR, xMaxR, yMaxR;
  double xMinP, yMinP, xMaxP, yMaxP;
  
  /* Polygone ferm ? */
  if (PLtestClosedPolygon(nbPts, poly))
    nbPts--;

  /* Polygone a un point !? */
  if (nbPts == 1)
    return (PLpointInPolygon(poly, 4, rectangle));

  /* Rectangles ex-inscrits */
  PLpolygonMinMax(4, rectangle, &xMinR, &yMinR, &xMaxR, &yMaxR);
  PLpolygonMinMax(nbPts, poly, &xMinP, &yMinP, &xMaxP, &yMaxP);

  /* Tester si les rectangles ex-inscrits s'intersectent */
  if ((xMinR > xMaxP) || (yMinR > yMaxP)
      || (xMaxR < xMinP) || (yMaxR < yMinP))
    return(FALSE);
  
  /* Regarder si un coin du rectangle est dans le polygone */
  if (nbPts > 2)
    for (i=0; i<4; i++) 
      if (PLpointInPolygon(&rectangle[2*i], nbPts, poly))
	return(TRUE);
  
  /* Regarder l'intersection des segments */
  for (i=0; i<3; i++) {
    for (j=0; j<nbPts-1; j++) 
      if (PLtestIntersectSegments(&rectangle[2*i], &rectangle[2*(i+1)], &poly[2*j], &poly[2*(j+1)]))
	return(TRUE);
    if (PLtestIntersectSegments(&rectangle[2*i], &rectangle[2*(i+1)], &poly[2*j], &poly[0]))
	return(TRUE);
  }
  for (j=0; j<nbPts-1; j++) 
    if (PLtestIntersectSegments(&rectangle[2*3], &rectangle[0], &poly[2*j], &poly[2*(j+1)]))
      return(TRUE);
  if (PLtestIntersectSegments(&rectangle[2*3], &rectangle[0], &poly[2*j], &poly[0]))
    return(TRUE);
  
  return(FALSE);
	  
}


/*----------------------------------------------------------------------
 * PLpolygonInsidePolygon   - 
 *
 * Description : Test if the polygon poly1 is inside the polygon poly2
 * Input       : 2 polygons
 * Returns     : TRUE if entirely inside, FALSE otherwise
 */
int PLpolygonInsidePolygon (int nbPts1, double *poly1, 
			    int nbPts2, double *poly2)
{
  int i, j;
  double xMinR, yMinR, xMaxR, yMaxR;
  double xMinP, yMinP, xMaxP, yMaxP;
  
  /* Polygones ferms ? */
  if (PLtestClosedPolygon(nbPts1, poly1))
    nbPts1--;
  if (PLtestClosedPolygon(nbPts2, poly2))
    nbPts2--;

  /* Test validite */
  if (nbPts1 < 1 || nbPts2 < 3) {
    printf ("PLpolygonInsidePolygon: unvalid parameters\n");
    return FALSE;
  }

  /* Polygone a un point !? */
  if (nbPts1 == 1)
    return (PLpointInPolygon(poly1, nbPts2, poly2));

  /* Rectangles ex-inscrits aux polygones */
  PLpolygonMinMax(nbPts1, poly1, &xMinR, &yMinR, &xMaxR, &yMaxR);
  PLpolygonMinMax(nbPts2, poly2, &xMinP, &yMinP, &xMaxP, &yMaxP);

  /* Test si rectangle ex-inscrit de poly1 est inclus dans celui de poly2 */ 
  if ((xMaxR > xMaxP) || (yMaxR > yMaxP)
      || (xMinR < xMinP) || (yMinR < yMinP))
    return(FALSE);
  
  /* Regarder si tous les sommets du polygone1 sont dans dans le polygone2 */
  for (i=0; i<nbPts1; i++) 
    if (!PLpointInPolygon(&poly1[2*i], nbPts2, poly2))
      return(FALSE);
  
  /* Regarder l'intersection des segments */
  for (i=0; i<nbPts1-1; i++) {
    for (j=0; j<nbPts2-1; j++) 
      if (PLtestIntersectSegments(&poly1[2*i], &poly1[2*(i+1)], &poly2[2*j], &poly2[2*(j+1)]))
	return FALSE;
    if (PLtestIntersectSegments(&poly1[2*i], &poly1[2*(i+1)], &poly2[2*j], &poly2[0]))
	return FALSE;
  }
  for (j=0; j<nbPts2-1; j++) 
    if (PLtestIntersectSegments(&poly1[2*i], &poly1[0], &poly2[2*j], &poly2[2*(j+1)]))
      return FALSE;
  if (PLtestIntersectSegments(&poly1[2*i], &poly1[0], &poly2[2*j], &poly2[0]))
    return FALSE;
  
  return TRUE;
	  
}


/*----------------------------------------------------------------------
 * PLrectangleInsidePolygon   - 
 *
 * Description : Test if a rectangle is inside a polygon
 * Input       : 1 polygon with 4 vertices and 1 general polygon
 * Returns     : TRUE if entirely inside, FALSE otherwise
 */
int PLrectangleInsidePolygon (double rectangle[2*4], int nbPts, double *poly)
{
  int i, j;
  double xMinR, yMinR, xMaxR, yMaxR;
  double xMinP, yMinP, xMaxP, yMaxP;
  
  /* Polygone ferm ? */
  if (PLtestClosedPolygon(nbPts, poly))
    nbPts--;

  /* Polygone doit avoir au moins 3 points */
  if (nbPts < 3)
    return (FALSE);

  /* Rectangles ex-inscrits */
  PLpolygonMinMax(4, rectangle, &xMinR, &yMinR, &xMaxR, &yMaxR);
  PLpolygonMinMax(nbPts, poly, &xMinP, &yMinP, &xMaxP, &yMaxP);

  /* Tester si le rectangle est inclus dans le rectangle polygone */
  if ((xMaxR > xMaxP) || (xMinR < xMinP)
      || (yMaxR > yMaxP) || (yMinR < yMinP))
    return FALSE;
  
  /* Regarder si les 4 coins du rectangle sont dans le polygone */
  for (i=0; i<4; i++) 
    if (!PLpointInPolygon(&rectangle[2*i], nbPts, poly))
      return(FALSE);
  
  /* Regarder l'intersection des segments */
  for (i=0; i<3; i++) {
    for (j=0; j<nbPts-1; j++) 
      if (PLtestIntersectSegments(&rectangle[2*i], &rectangle[2*(i+1)], &poly[2*j], &poly[2*(j+1)]))
	return(FALSE);
    if (PLtestIntersectSegments(&rectangle[2*i], &rectangle[2*(i+1)], &poly[2*j], &poly[0]))
	return(FALSE);
  }
  for (j=0; j<nbPts-1; j++) 
    if (PLtestIntersectSegments(&rectangle[2*3], &rectangle[0], &poly[2*j], &poly[2*(j+1)]))
      return(FALSE);
  if (PLtestIntersectSegments(&rectangle[2*3], &rectangle[0], &poly[2*j], &poly[0]))
    return(FALSE);
  
  return(TRUE);
}

/*----------------------------------------------------------------------
 * PLpointInPolygon   - 
 *
 * Description : Test if a point is inside a polygon
 * Input       : 1 point and 1 general polygon
 * Returns     : 1 if inside or 2 on the edge, FALSE (0) otherwise
 *
 * Comment : Based on the number of intersections of an horizontal ray
 *           that passes througt the point with the polygon edges.
 *           Twice faster than PLpointInPolygonOld.
 *           
 */

int PLpointInPolygon (double point[2], int nbPts, double *poly)
{
  int i;           /* point index */
  double x;        /* x intersection of e with ray */
  int Rcross = 0;  /* number of right edge/ray crossing */
  int Lcross = 0;  /* number of left edge/ray crossing */
  int Rstrad, Lstrad;  /* flags indicating the edge strads the x axis */
  double px1, py1;     /* coordinates of point index i-1 of poly */
  double px2, py2;     /* coordinates of point index i of poly */

  /* fist point of first edge (ie, last point of poly) */
  i=nbPts-1;
  px1 = poly[2*i]   - point[0];
  py1 = poly[2*i+1] - point[1];

  /* First check if q=(0,0) is a vertex */
  if (px1 == 0 && py1 == 0)
    return 2;

  /* for each edge e = (i-1, i), see if crosses rays */
  for (i=0; i<nbPts; i++) {
    
    /* second edge point */
    px2 = poly[2*i]   - point[0];
    py2 = poly[2*i+1] - point[1];
    
    /* First check if q=(0,0) is a vertex */
    if (px2 == 0 && py2 == 0)
      return 2;

    /* check if e straddles x axis, with bias above/below */
    Rstrad = (py1 > 0) != (py2 > 0);
    Lstrad = (py1 < 0) != (py2 < 0);
    if (Rstrad || Lstrad) {

      /* compute intersection of e with x axis */
      x = (px2 * py1 - px1 * py2) / (py1 - py2);

      if (Rstrad && x > 0.0) Rcross++;
      if (Lstrad && x < 0.0) Lcross++;
    } /* end straddle computation */

    /* next edge */
    px1 = px2;
    py1 = py2;

  } /* end for */

  /* q on an edge if L/Rcross counts are not the same parity */
  if ((Rcross%2) != (Lcross%2))
    return 2;
  
  /* q inside iff an odd number of crossing */
  if ((Rcross%2) == 1)
    return 1;
  
  /* outside */
  else return FALSE;
}

/*----------------------------------------------------------------------
 * PLpointInPolygonOld   - 
 *
 * Description : Test if a point is inside a polygon
 * Input       : 1 point and 1 general polygon
 * Returns     : TRUE if inside or on the edge, FALSE otherwise
 */

int PLpointInPolygonOld (double point[2], int nbPts, double *poly)
{
  int i, indx, indy;
  double sum=0.;
  double dx, dy;
  double angle, prevAngle, angle0, dAng;
  
  /* Angle avec le premier point */
  dx = poly[0] - point[0];
  dy = poly[1] - point[1];
  if (fabs(dx) < 0.00001 && fabs(dy) < 0.00001)
    return(TRUE);
  angle0 = prevAngle = atan2(dy, dx);
  
  /* Somme des differences des angles */
  for (i=0; i<nbPts; i++) {
    indx = 2*i;
    indy = 2*i+1;
    dx = poly[indx] - point[0];
    dy = poly[indy] - point[1];
    if (fabs(dx) < 0.00001 && fabs(dy) < 0.00001)
      return(TRUE);
    angle = atan2(dy, dx);
    dAng = angle - prevAngle;
    if (dAng > M_PI) dAng -= 2.*M_PI;
    else if (dAng < -M_PI) dAng += 2.*M_PI;
    sum += dAng;
    prevAngle = angle;
  }

  /* Et entre le dernier et le premier point */
  dAng = angle0 - prevAngle;
  if (dAng > M_PI) dAng -= 2.*M_PI;
  else if (dAng < -M_PI) dAng += 2.*M_PI;
  sum += dAng;
  
  /* sum = 2PI : dedans , sum = 0: dehors */
  if (fabs(sum) < M_PI)
    return(FALSE);
  return(TRUE);
}


/*----------------------------------------------------------------------
 * PLpointSegDist
 *
 * Description : mputes dist between a point and a segment
 * Input       : 1 point and 1 segment
 * Returns     : distance
 */

double PLpointSegDist (double point[2], double *seg)
{
  double vx, vy;
  double d1, d2, dist;
  double dx1, dx2, dy1, dy2;

  /* vector of the segment */
  vx = seg[2] - seg[0];
  vy = seg[3] - seg[1];

  /* vectors P1P and P2P */
  dx1 = point[0] - seg[0];
  dy1 = point[1] - seg[1];
  dx2 = point[0] - seg[2];
  dy2 = point[1] - seg[3];

  /* test if the perpendicular projection of the point is on the segment */

  /* distance  between P1 and the projection of P on S 
     (scalar product P1P2.P1P) */
  d1 = dx1*vx + dy1*vy;
  
  /* distance  between P2 and the projection of P on S 
     (scalar product P1P2.P2P)*/
  d2 = dx2*vx + dy2*vy;
  
  /* proj(P) on S is not in [P1,P2] */
  if ( d1 <= 0 || d2 >= 0) {

    /* the minimum distance is one of the 2 extremities */
    return MIN(hypot(dx1, dy1), hypot(dx2, dy2));
  }

  /* proj on the seg : the mini dist is the vecorial product P1P2^P1P */
  dist = (vx*dy1 - vy*dx1) / sqrt(vx*vx + vy*vy);
  
#if DEBUG>=1
  printf ("min seg dist = %g \n", dist);
#endif
  return fabs(dist); 

}

/*----------------------------------------------------------------------
 * PLprojPointOnSeg
 *
 * Description : Computes projection of a pt on a line defined by a seg.
 * Input       : 1 point, 1 segment and wanted dist between seg and 
 *               "proj pt" (0 for a point on the seg)
 * Returns     : 0 if proj on seg, 1 if outside ext1, 2 if outside ext2
 *               and the point proj on the line.
 */

int PLprojPointOnSeg (double pt[2], double *seg, 
		      double requiredDistToSeg, double *proj)
{
  double d1q, d2q, dpq;
  double v12[2], v1p[2], v2p[2];
  double n;
  double q[2];

  /* vector of the segment */
  v12[0] = seg[2] - seg[0];
  v12[1] = seg[3] - seg[1];

  /* norme of seg */
  n = hypot(v12[0], v12[1]);
  if ( n < 0.000001) return 0;

  /* normalisation of the vector of the segment */
  v12[0] /= n;
  v12[1] /= n;

  /* vectors P1P and P2P */
  v1p[0] = pt[0] - seg[0];
  v1p[1] = pt[1] - seg[1];
  v2p[0] = pt[0] - seg[2];
  v2p[1] = pt[1] - seg[3];

  /* test if the perpendicular projection of the point is on the segment */

  /* distance  between P1 and Q, the projection of P on S 
     (scalar product P1P.P1P2) */
  d1q = v1p[0]*v12[0] + v1p[1]*v12[1];
  
  /* distance  between P2 and Q, the projection of P on S 
     (scalar product P2P.P1P2)*/
  d2q = v2p[0]*v12[0] + v2p[1]*v12[1];

  /* coord of proj(P) on S */
  q[0] = seg[0] + d1q * v12[0];
  q[1] = seg[1] + d1q * v12[1];
  
  /* Required point is just the proj of P on seg */
  if (requiredDistToSeg == 0) {
    proj[0] = q[0];
    proj[1] = q[1];
  }

  /* New point distant from requiredDistToSeg */
  else {
    /* PQ dist is the vecorial product P1P2^PP1 */
    dpq = - v12[0]*v1p[1] + v12[1]*v1p[0];
  
    /* proj(P)Q / distRequired  = Pproj(P) / dpq */
    proj[0] = requiredDistToSeg / dpq * (q[0] - pt[0]) +  q[0];
    proj[1] = requiredDistToSeg / dpq * (q[1] - pt[1]) +  q[1];
  }

  /* proj(P) on S is not in [P1,P2] but near P1 */
  if ( d1q <= 0)
    return 1;

  /* proj(P) on S is not in [P1,P2] but near P2 */
  if (d2q >= 0) 
    return 2;

  /* proj(P) is in [p1,p2] */
  return 0;
}

/*----------------------------------------------------------------------
 * PLpointNearestPolygonEdge
 *
 * Description : Look for nearest edge of a polygon from a given point
 * Input       : 1 point and 1 general polygon
 * Returns     : index i of nearest edge [Pi, Pi+1]
 */

int PLpointNearestPolygonEdge (double point[2], int nbPts, double *poly,
			       double *minDist)
{
  int i;
  double dist, min=100000;
  int nearest;
  double last[4];
  
  /* dist for every edge */
  for (i=0; i<nbPts-1; i++) {
    dist = PLpointSegDist(point, &poly[2*i]);
    if (dist < min) {
      min = dist;
      nearest = i;
    }
  }

  /* last edge */
  last[0] = poly[2*i];
  last[1] = poly[2*i+1];
  last[2] = poly[0];
  last[3] = poly[1];

  dist = PLpointSegDist(point, last);
  if (dist < min) {
    min = dist;
    nearest = i;
  }
  
  /* result */
  *minDist = min;
  return nearest;
}

/*----------------------------------------------------------------------
 * PLtestIntersectSegments   - 
 *
 * Description : Test segments intersection
 * Input       : 4 points: 2 for the 1st segment and 2 for the 2nd one
 * Returns     : TRUE if the intersect, FALSE otherwise
 */

int PLtestIntersectSegments (double p1s1[2], double p2s1[2],
			     double p1s2[2], double p2s2[2])
{
  double dx, dy;
  double p1, p2;

  /* s2 intersect la droite d1 ? */
  /* Vecteur P1-P2 de S1 */
  dx = p2s1[0] - p1s1[0]; 
  dy = p2s1[1] - p1s1[1]; 

  /* Produit vectoriel P1s1-P2s1 * P1s1-P1s2 */
  p1 = dx * (p1s2[1] - p1s1[1]) - dy * (p1s2[0] - p1s1[0]);
  /* Produit vectoriel P1s1-P2s1 * P1s1-P2s2 */
  p2 = dx * (p2s2[1] - p1s1[1]) - dy * (p2s2[0] - p1s1[0]);

  /* Les deux points de s2 sont du meme cote de d1 */
  if (p1 * p2 > 0.)
    return(FALSE);

  /* s1 intersect d2 ? */
  dx = p2s2[0] - p1s2[0]; 
  dy = p2s2[1] - p1s2[1]; 

  p1 = dx * (p1s1[1] - p1s2[1]) - dy * (p1s1[0] - p1s2[0]);
  p2 = dx * (p2s1[1] - p1s2[1]) - dy * (p2s1[0] - p1s2[0]);

  /* Les deux points de s2 sont du meme cote de d1 */
  if (p1 * p2 > 0.)
    return(FALSE);

  return(TRUE); 
}

/*----------------------------------------------------------------------
 * PLintersectSegments   - 
 *
 * Description : Computes segments intersection
 * Input       : 4 points: 2 for the 1st segment and 2 for the 2nd one
 * Output      : 1 point: the intersection if any
 * Returns     : TRUE if the intersect, FALSE otherwise
 */
int PLintersectSegments (double p1s1[2], double p2s1[2],
			 double p1s2[2], double p2s2[2],
			 double inter[2])
{
  double det, alpha, beta;
  double x1,x2,x3,x4;
  double y1,y2,y3,y4;

  x1 = p1s1[0];
  y1 = p1s1[1];
  x2 = p2s1[0];
  y2 = p2s1[1];
  x3 = p1s2[0];
  y3 = p1s2[1];
  x4 = p2s2[0];
  y4 = p2s2[1];

  /* Produit vectoriel */
  det = (x1 - x2)*(y3 - y4) + (x3 - x4)*(y2 - y1);
  /* Les droites sont paralleles */
  if (fabs(det) < 1e-4)
    return 0;

  alpha = -(y2*(-x3+x4) + y3*(x2-x4) + y4*(x3-x2)) / det;
  /* L'intersection des droites n'est pas sur le segment S1 */
  if (alpha < 0 || alpha > 1.0)
    return 0;

  beta  = (x1* (y2-y4) - x2*(y1 - y4) - x4*(y2 - y1))/det; 
  /* L'intersection des droites n'est pas sur le segment S2 */
  if (beta < 0 || beta > 1.0)
    return 0;

  /* Point d'intersection */
  inter[0] = alpha*x1 + (1-alpha)*x2;
  inter[1] = alpha*y1 + (1-alpha)*y2;
  return 1;
}

/*----------------------------------------------------------------------
 * PLbissectorPt
 *
 * Description : computes a point on a bissector at a given distance
 *               from the vertice and/or from the segs
 * Input       : 3 points (p1 is the vertice) and 1 required distance
 * Outputs     : the point at dist from vertice (if non NULL) and
 *             : the point at dist from segs (if non NULL) 
 * Returns     : TRUE
 */

int PLbissectorPt(double *p0, double *p1, double *p2,
		  double dist, double *fromVertice, 
		  double *fromSeg)
{
   double dx0,dy0, dx2,dy2, d, bsin, bx,by;

   /* normalize segments direction */
   d = hypot(p0[0]-p1[0], p0[1]-p1[1]);
   dx0 = (p0[0]-p1[0])/d;
   dy0 = (p0[1]-p1[1])/d;
   
   d = hypot(p2[0]-p1[0], p2[1]-p1[1]);
   dx2 = (p2[0]-p1[0])/d;
   dy2 = (p2[1]-p1[1])/d;
   
   /* compute new point at dist from segs */
   if (fromSeg) {

     /* find direction and norm of bissector */
     bsin = dx2*dy0 - dy2*dx0;
   
     /* save result */
     if (fabs(bsin)<1e-5) {
       /* (0,1) and (1,2) are parallel */
       bx = -dy0;
       by = dx0;
     } 
     else {
       bx = -(dx0+dx2)/bsin;
       by = -(dy0+dy2)/bsin;
     }
     fromSeg[0] = p1[0] - dist*bx;
     fromSeg[1] = p1[1] - dist*by;
   }

   /* computes new point at dist from vertice */
   if (fromVertice) {
     d = hypot(dx0+dx2, dy0+dy2);
     if (d < 1e-5) {
       bx = -dy0;
       by = dx0;
     }
     else {
       bx = -(dx0+dx2)/d;
       by = -(dy0+dy2)/d;
     }
     fromVertice[0] = p1[0] - dist * bx;
     fromVertice[1] = p1[1] - dist * by;
   }
   return 1;
}

/*----------------------------------------------------------------------
 * PLgrowSeg
 *
 * Description : Computes a seg parallel to a gievn seg at a given signed
 *               distance. Usefull to grow a segment to make a rectangle.
 *               For a positive input distance, the new seg
 *               will on the left for someone walking from pt1 to pt2.
 * Input       : 1 input seg, the distance, 1 output seg
 * Returns     : TRUE 
 */

int 
PLgrowSeg(double *src, double dist, double *dst)
{
  double angle;
  double cs, sn;
  double dx, dy;

  /* Pour utiliser dessin_rects il faut calculer  les coordonnees
     des rectangles correspondant aux segements grossis cote matiere 
     (xMapObst et yMapObst) */
  angle = atan2(src[3]-src[1], src[2]-src[0]) - M_PI/2.;
  cs = cos(angle);
  sn = sin(angle);
  
  dx = dist * cos(angle);
  dy = dist * sin(angle);
  
  dst[0] = src[0] + dx;
  dst[1] = src[1] + dy;
  dst[2] = src[2] + dx;
  dst[3] = src[3] + dy;
  
  return 1;
}

/*----------------------------------------------------------------------
 * PLgrowPolygon
 *
 * Description : Make a polygon bigger (or smaller if dist < 0)
 *               For a positive input distance, the new edges
 *               will on the left for someone walking from pt1 to pt2.
 *               In other word, a clock wise polygon will be enlarged
 *               with dist > 0.
 *               polygon must be given in clock wise.
 * Input       : input poly, dist, output poly (same size)
 * Returns     : TRUE 
 */
int
PLgrowPolygon(int nbPts, double *src, double dist, double *dst)
{
   int i;
   
   /* bad parameters */
   if (nbPts < 2) {
     fprintf(stderr, "PLgrowPolygon: bad nbPts %d\n", nbPts);
     return 0;
   }

   /* just a segment */
   if (nbPts == 2) {
     return PLgrowSeg(src, dist, dst);
   }

   /* compute each points */
   for(i=0; i<nbPts; i++) {
      PLbissectorPt(&src[2*((i+nbPts-1)%nbPts)],
		    &src[2*i],
		    &src[2*((i+1)%nbPts)],
		    dist,
		    NULL,
		    &dst[2*i]);
   }

   return 1;
}

/*----------------------------------------------------------------------
 * PLpolygonMinMax   - 
 *
 * Description : Computes min, max polygon coordinates
 * Input       : 1 polygon 
 * Output      : the 4-th coordinates
 * Returns     : void
 */

void PLpolygonMinMax(int nbPts, double *poly,
		     double *xMin, double *yMin, double *xMax, double *yMax)
{
  int i, indx, indy;
  double minX;
  double minY;
  double maxX;
  double maxY;

  minX = maxX = poly[0];
  minY = maxY = poly[1];

  for(i=1; i<nbPts; i++) {
    indx = 2*i;
    indy = 2*i+1;
    if (poly[indx] < minX) minX = poly[indx];
    else if (poly[indx] > maxX) maxX = poly[indx];
    if (poly[indy] < minY) minY = poly[indy];
    else if (poly[indy] > maxY) maxY = poly[indy];
  }
  *xMin = minX;
  *xMax = maxX;
  *yMin = minY;
  *yMax = maxY;
}


/*----------------------------------------------------------------------
 * PLtestClosedPolygon   - 
 *
 * Description : Test if a polygon is closed, 
 *               ie, if 1st and last points are identical
 * Input       : 1 polygon
 * Returns     : TRUE if it is closed, FALSE otherwise
 */
int PLtestClosedPolygon(int nbPts, double *poly)
{
  /* Polygone ferm ? */
  if (nbPts>1)
    if (fabs(poly[0]-poly[2*(nbPts-1)])<1e-6 &&
	fabs(poly[1]-poly[2*(nbPts-1)+1])<1e-6)
      return 1;
  
  return 0;
}

/*----------------------------------------------------------------------
 * PLprintPolygon
 *
 * Description : Print a polygon
 * Input       : 1 filedescripor and 1 polygon
 * Returns     : TRUE if it is closed, FALSE otherwise
 */
void PLprintPolygon(FILE *pfile, int nbPts, double *poly, int fullDataFlag)
{
  int i;

  if (fullDataFlag) {
    if (PLtestClosedPolygon(nbPts, poly))
      fprintf(pfile, "Closed polygon of %d points:\n", nbPts);
    else
      fprintf(pfile, "Open polygon of %d points:\n", nbPts);
  }
  for(i=0; i<nbPts; i++) {
    if (fullDataFlag)
      fprintf(pfile, "%3d: ", i);
    fprintf(pfile, "  %10f %10f\n", poly[2*i], poly[2*i+1]);
  }
}


/*----------------------------------------------------------------------
 * PLscanOneH2Polygon
 *
 * Description : Scan polygon of type "H2" one by one
 *               H2 obstacles file format :
 *
 *                 [any comments...]
 *                 OBSTACLES  [any comments...]
 *                 x0  y0
 *                 x1  y1
 *                 [...]
 *                 END     [any comments...]
 *                 x0  y0
 *                 x1  y1
 *                 [...]
 *                 END    ([any comments...]
 *                 [and so on ...]
 *
 *               If scan is OK the polygon is allocated by this function.
 *               See testPolygons.c for an example
 *
 * Input       : 1 file descriptor 
 *               1 reference on the number of points of the read polygon
 * Returns     : reference on the polygon or NULL
 */
double * PLscanOneH2Polygon(FILE *pfile, int *nbPts)
{
  int i, n;
  double x, y;
  double *newpoly=NULL;
  char string[128];
  char c;

  *nbPts = 0;

  /* File header 'OBSTACLES' or indication of end of previous obstacle 'END' */
  do {
    n = fscanf(pfile, "%s", string);
    if (n == 1 && 
	(strcmp("OBSTACLES", string) == 0 || strcmp("END", string) == 0)) 
      break;
    if (n == EOF) {
      printf ("PLscanOneH2Polygon: bad obstacle format\n");
      return 0;
    }
  } while(1);

  /* Read until the end of the line (for eventual comments) */
  do {
    c = getc(pfile);
  } while (c!='\n');

  /* The polygon */
  i = 0;
  do {
    n = fscanf(pfile, "%lf %lf", &x, &y);

    /* No (more?) coordinates to be read */
    if (n != 2)
      break;

    newpoly = (double *)realloc(newpoly, 2*(i+1)*sizeof(double));
    newpoly[2*i] = x;
    newpoly[2*i+1] = y;
    i++;
  } while(1);
  
  /* Every thing is OK */
  *nbPts = i;
  return newpoly;
}


/*----------------------------------------------------------------------
 * PLpointInHorizontalRectangle   - 
 *
 * Description : Test if a point is inside an horizontal rectangle
 * Input       : 1 point and 1 rectangle xmin, ymin, xmax, ymax
 * Returns     : TRUE if inside or on the edge, FALSE otherwise
 */

int PLpointInHorizontalRectangle (double point[2], double rect[2*2])
{
  if (point[0] >= rect[0] && point[0] <= rect[2] &&
      point[1] >= rect[1] && point[1] <= rect[3]) 
    return TRUE;
  return FALSE;
}

/*----------------------------------------------------------------------
 * PLclipSegment   - 
 *
 * Description : Test if a point is inside an horizontal rectangle
 * Input       : Extremities ptSeg1 and ptSeg2 of the segment
 *               1 rectangle xmin, ymin, xmax, ymax
 *               Extremities pt1out and pt2out of clipped segment
 * Returns     : 0 if the segment is outside the rectangle, 
 *               1 if the segment is inside the rectangle, 
 *               2 if the segment has been clipped
 */

int PLclipSegment(double ptSeg1[2], double ptSeg2[2], 
		  double rect[2*2],
		  double pt1out[2], double pt2out[2])
{
  double dist1, dist2;
  double pt1[2];
  double pt2[2];
  double pt3[2];
  double pt4[2];
  double ptI[2];
  double ptI2[2];
  int numPtI=0;
  double *ptIPtr;

  pt1out[0] = ptSeg1[0];
  pt1out[1] = ptSeg1[1];
  pt2out[0] = ptSeg2[0];
  pt2out[1] = ptSeg2[1];

  /* XXX  (redondant avec ce qui suit)
     les 2 extremites du segments sont hors des limites ET du mme cot */
  if ((ptSeg1[0] < rect[0] && ptSeg1[2] < rect[0]) ||
      (ptSeg1[0] > rect[2] && ptSeg1[2] > rect[2]) ||
      (ptSeg1[1] < rect[1] && ptSeg1[3] < rect[1]) ||
      (ptSeg1[1] > rect[3] && ptSeg1[3] > rect[3])) {
    return 0;
  }

  /* les 2 extremites du segments sont dans les limites */
  if (PLpointInHorizontalRectangle(ptSeg1, rect) && 
      PLpointInHorizontalRectangle(ptSeg2, rect)) {
    return 1;
  }

  pt1[0] = rect[0];
  pt1[1] = rect[1];
  pt2[0] = rect[0];
  pt2[1] = rect[3];
  pt3[0] = rect[2];
  pt3[1] = rect[3];
  pt4[0] = rect[2];
  pt4[1] = rect[1];

  /**
   **  Le point 1 deborde mais le point 2 est  l'interieur 
   **/
  if (PLpointInHorizontalRectangle(ptSeg2, rect)) {
    
    /* On cherche le cote qui intersect le segment */
    if (!PLintersectSegments(pt1,pt2,ptSeg1,ptSeg2,ptI))
      if (!PLintersectSegments(pt2,pt3,ptSeg1,ptSeg2,ptI))
	if (!PLintersectSegments(pt3,pt4,ptSeg1,ptSeg2,ptI))
	  if (!PLintersectSegments(pt4,pt1,ptSeg1,ptSeg2,ptI))
	    printf ("seglocClipMap : no intersection ?! xm %f ym %f xM %f yM %f \n",
		    ptSeg1[0], ptSeg1[1], ptSeg2[0], ptSeg2[1]);
    
    /* On coupe le point 1 du segment */
    pt1out[0] = ptI[0];
    pt1out[1] = ptI[1];
    return 2;
  }
  
  /**
   ** Le point 2 deborde mais le point 1 est  l'interieur 
   **/
  if (PLpointInHorizontalRectangle(ptSeg1, rect)) {
    
    /* On cherche le cote qui intersect le segment */
    if (!PLintersectSegments(pt1,pt2,ptSeg1,ptSeg2,ptI)) 
      if (!PLintersectSegments(pt2,pt3,ptSeg1,ptSeg2,ptI))
	if (!PLintersectSegments(pt3,pt4,ptSeg1,ptSeg2,ptI))
	  if (!PLintersectSegments(pt4,pt1,ptSeg1,ptSeg2,ptI))
	    printf ("seglocClipMap : no intersection ?! xm %f ym %f xM %f yM %f \n",
		    ptSeg1[0], ptSeg1[1], ptSeg2[0], ptSeg2[1]);
    
    /* On coupe le point 2 du segment */
    pt2out[0] = ptI[0];
    pt2out[1] = ptI[1];
    return 2;
  }
  
  /**
   ** les 2 points sont  l'exterieur 
   **/

  numPtI = 0;
  ptIPtr = ptI;
  /* Test intersection avec pt1-pt2 */
  if (PLintersectSegments(pt1,pt2,ptSeg1,ptSeg2,ptIPtr)) {
    numPtI++;
    ptIPtr = ptI2;
  }
  /* Test intersection avec pt2-pt3 */
  if (PLintersectSegments(pt2,pt3,ptSeg1,ptSeg2,ptIPtr)) {
    numPtI++;
    ptIPtr = ptI2;
  }
  /* on n'a trouv qu' 1 intersection: on continue */
  if (numPtI < 2) {
    /* Test intersection avec pt3-pt4 */
    if (PLintersectSegments(pt3,pt4,ptSeg1,ptSeg2,ptIPtr)) {
      numPtI++;
      ptIPtr = ptI2;
    }
  }
  
  /* on n'a trouv qu' 1 intersection: on continue encore */
  if (numPtI < 2) {
    /* Test intersection avec pt4-pt1 */
    if (PLintersectSegments(pt4,pt1,ptSeg1,ptSeg2,ptIPtr)) {
      numPtI++;
      ptIPtr = ptI2;
    }
  }
    
  /* Pas d'intersection */
  if (numPtI==0) {
    return 0;
  }

  /* Double intersection : on coupe les deux extrmits */
  else if (numPtI==2) {
	
    dist1 = sqrt(SQR(ptI[0]-pt1out[0])  + SQR(ptI[1]-pt1out[1]));
    dist2 = sqrt(SQR(ptI2[0]-pt1out[0]) + SQR(ptI2[1]-pt1out[1]));
	
    /* C'est ptI qui est plus proche de sinput->first */
    if (dist1 < dist2) {
      
      /* On coupe le point 1 du segment */
      pt1out[0] = ptI[0];
      pt1out[1] = ptI[1];
      
      /* On coupe le point 2 du segment */
      pt2out[0] = ptI2[0];
      pt2out[1] = ptI2[1];
    }
    /* C'est ptI2 qui est plus proche de sinput->first */
    else {
      /* On coupe le point 1 du segment */
      pt1out[0] = ptI2[0];
      pt1out[1] = ptI2[1];
      
      /* On coupe le point 2 du segment */
      pt2out[0] = ptI[0];
      pt2out[1] = ptI[1];
    }
  } /* double intersection */

  else {
    printf ("seglocClipMap: %d intersections ??? \n", numPtI);
  }
  return 2;
}

    
  
